#!/usr/bin/env python3

from collections.abc import Iterator, Mapping, Sequence
from typing import Literal

from .dfs import depth_first_search
from .johnson import johnson
from .modules import PyModule
from .tarjan import strongly_connected_components


def detect_cycles(
    strategy: Literal["dfs", "tarjan"],
    graph: Mapping[PyModule, Sequence[PyModule]],
) -> Iterator[tuple[PyModule, ...]]:
    if strategy == "dfs":
        return depth_first_search(graph)
    if strategy == "tarjan":
        return (scc for scc in strongly_connected_components(graph) if len(scc) > 1)
    if strategy == "johnson":
        return johnson(graph)
    raise NotImplementedError()



#!/usr/bin/env python3

from collections.abc import Iterator, Mapping, Sequence
from typing import TypeVar

from networkx import DiGraph, simple_cycles

from .type_defs import Comparable

T = TypeVar("T", bound=Comparable)


def johnson(graph: Mapping[T, Sequence[T]]) -> Iterator[tuple[T, ...]]:
    for sc in simple_cycles(DiGraph([(v, w) for v, vertices in graph.items() for w in vertices])):
        yield tuple(sc)



#!/usr/bin/env python3

import os
import time
from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from pathlib import Path

from .modules import PyModule


def scan_packages(packages: Sequence[Path]) -> Iterator[PyModule]:
    paths_to_ignore: set[Path] = set()
    for package_path in packages:
        for root, _dirs, files in os.walk(package_path):
            root_path = Path(root)

            if root_path.name.startswith(".") or root_path.name == "__pycache_":
                paths_to_ignore.add(root_path)
                continue

            if any(root_path.is_relative_to(p) for p in paths_to_ignore):
                continue

            for file in files:
                # May be a regular package or a module
                file_path = root_path / file
                try:
                    yield PyModule(package=package_path, path=file_path)
                except ValueError:
                    pass

            # May be a namespace package
            try:
                yield PyModule(package=package_path, path=root_path)
            except ValueError:
                pass


def parse_files(files: Sequence[str]) -> Iterator[PyModule]:
    cwd = Path.cwd()
    for file in files:
        raw_root_path, raw_rel_file_path = file.split("::", 1)

        root_path = Path(raw_root_path)
        if not root_path.is_absolute():
            root_path = cwd / raw_root_path

        rel_file_path = Path(raw_rel_file_path)
        yield PyModule(
            package=root_path / rel_file_path.parts[0],
            path=root_path / rel_file_path,
        )


@dataclass(frozen=True, kw_only=True)
class OutputsFilePaths:
    log: Path
    graph: Path


def get_outputs_file_paths(outputs_folder: Path | None, outputs_filename: str) -> OutputsFilePaths:
    if not outputs_folder:
        outputs_folder = Path.home() / Path(".local", "py-import-cycles", "outputs")
    outputs_folder.mkdir(parents=True, exist_ok=True)
    if not outputs_filename:
        outputs_filename = str(int(time.time()))
    return OutputsFilePaths(
        log=(outputs_folder / outputs_filename).with_suffix(".log"),
        graph=(outputs_folder / outputs_filename).with_suffix(".gv"),
    )



#!/usr/bin/env python3

import itertools
import random
import sys
from collections import defaultdict
from collections.abc import Iterator, Mapping, Sequence
from pathlib import Path
from typing import DefaultDict, Iterable, Literal, NamedTuple, TypeVar

from graphviz import Digraph

from .log import logger
from .modules import PyModule, PyModuleType
from .type_defs import Comparable


class ImportEdge(NamedTuple):
    title: str
    from_py_module: PyModule
    to_py_module: PyModule
    edge_color: str


T = TypeVar("T")
TC = TypeVar("TC", bound=Comparable)


def make_graph(
    filepath: Path,
    opt_strategy: Literal["dfs", "tarjan"],
    import_cycles: Sequence[tuple[PyModule, ...]],
) -> None:
    sys.stderr.write(f"Write graph data to {filepath}\n")

    if not (edges := _make_edges(opt_strategy, import_cycles)):
        logger.debug("No such edges for graph")
        return

    d = Digraph("unix", filename=filepath)

    with d.subgraph() as ds:
        for edge in edges:
            ds.node(
                str(edge.from_py_module.name),
                shape=_get_shape(edge.from_py_module),
            )

            ds.node(
                str(edge.to_py_module.name),
                shape=_get_shape(edge.to_py_module),
            )

            ds.attr("edge", color=edge.edge_color)

            if edge.title:
                ds.edge(
                    str(edge.from_py_module.name),
                    str(edge.to_py_module.name),
                    edge.title,
                )
            else:
                ds.edge(str(edge.from_py_module.name), str(edge.to_py_module.name))

    d.unflatten(stagger=50)
    d.view()


def _get_shape(py_module: PyModule) -> str:
    return "" if py_module.type is PyModuleType.MODULE else "box"


def pairwise(iterable: Iterable[T]) -> Iterator[tuple[T, T]]:
    # pairwise('ABCDEFG') --> AB BC CD DE EF FG
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def dedup_edges(cycles: Iterable[tuple[TC, ...]]) -> Sequence[tuple[TC, TC]]:
    edges: set[tuple[TC, TC]] = set()
    for cycle in cycles:
        edges |= frozenset(pairwise(cycle))
    return list(edges)


def count_degree(edges: Sequence[tuple[T, T]]) -> Mapping[T, int]:
    degrees: DefaultDict[T, int] = defaultdict(int)
    for edge in edges:
        degrees[edge[0]] -= 1
        degrees[edge[1]] += 1
    return degrees


def badness(edges: Sequence[tuple[T, T]]) -> Mapping[tuple[T, T], float]:
    # Get some idea of the badness of an edge by computing
    # the average of the degrees of its vertices.
    degree = count_degree(edges)
    return {edge: 0.5 * (degree[edge[0]] + degree[edge[1]]) for edge in edges}


def normalize(value: float, lower: float, higher: float) -> float:
    if not lower <= value < higher:
        raise ValueError(f"{value} is not within bounds")
    if higher < lower:
        raise ValueError("bounds are reverted")
    if higher == lower:
        raise ValueError("must use different bounds")
    return (value - lower) / (higher - lower)


def _make_edges(
    opt_strategy: Literal["dfs", "tarjan"],
    import_cycles: Sequence[tuple[PyModule, ...]],
) -> Sequence[ImportEdge]:
    if opt_strategy == "dfs":
        return _make_dfs_import_edges(import_cycles)
    return _make_only_cycles_edges(import_cycles)


def _make_dfs_import_edges(
    cycles: Sequence[tuple[PyModule, ...]],
) -> Sequence[ImportEdge]:
    edges = badness(dedup_edges(cycles))
    if not (edges_values := edges.values()):
        return []

    bound = max(abs(min(edges_values)), abs(max(edges_values)))

    out = []
    for edge, value in edges.items():
        nn = int(255 * normalize(value, -bound, bound + 1))
        assert 0 <= nn < 256
        if value < 0:
            color = "#{:02x}{:02x}{:02x}".format(  # pylint: disable=consider-using-f-string
                0, 255 - nn, nn
            )
        else:
            color = "#{:02x}{:02x}{:02x}".format(  # pylint: disable=consider-using-f-string
                nn, 0, 255 - nn
            )
        out.append(ImportEdge(str(value), edge[0], edge[1], color))
    return out


def _make_only_cycles_edges(
    import_cycles: Sequence[tuple[PyModule, ...]],
) -> Sequence[ImportEdge]:
    edges: set[ImportEdge] = set()
    for nr, import_cycle in enumerate(import_cycles, start=1):
        color = "#{:02x}{:02x}{:02x}".format(  # pylint: disable=consider-using-f-string
            random.randint(50, 200),
            random.randint(50, 200),
            random.randint(50, 200),
        )

        start_py_module = import_cycle[0]
        for next_py_module in import_cycle[1:]:
            edges.add(
                ImportEdge(
                    f"{str(nr)} ({len(import_cycle) - 1})",
                    start_py_module,
                    next_py_module,
                    color,
                )
            )
            start_py_module = next_py_module
    return sorted(edges)



#!/usr/bin/env python3

from __future__ import annotations

from collections.abc import Sequence
from enum import auto, Enum
from pathlib import Path
from string import ascii_letters
from typing import Final


def _parse_part(part: str) -> str:
    # A variable name must start with a letter or the underscore character.
    # A variable name cannot start with a number.
    # A variable name can only contain alpha-numeric characters and underscores (A-z, 0-9, and _ )
    if not part:
        raise ValueError(part)
    # TODO remove "*"; at the moment this is handled later in "make_module_from_name"
    if part[0] not in ascii_letters + "_*":
        raise ValueError(part[0])
    return part


class ModuleName:
    """This class is inspired by pathlib.Path"""

    def __init__(self, *parts: str | ModuleName) -> None:
        self._parts: Final[tuple[str, ...]] = tuple(
            _parse_part(entry)
            for part in parts
            for entry in (part.parts if isinstance(part, ModuleName) else part.split("."))
        )

    def __str__(self) -> str:
        return ".".join(self._parts)

    def __hash__(self) -> int:
        return hash(self._parts)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ModuleName):
            return NotImplemented
        return self._parts == other._parts

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, ModuleName):
            return NotImplemented
        return self._parts < other._parts

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, ModuleName):
            return NotImplemented
        return self._parts > other._parts

    def __le__(self, other: object) -> bool:
        return self < other or self == other

    def __ge__(self, other: object) -> bool:
        return self > other or self == other

    @property
    def parts(self) -> Sequence[str]:
        return self._parts

    @property
    def parent(self) -> ModuleName:
        return ModuleName(*self._parts[:-1])

    def joinname(self, *names: str | ModuleName) -> ModuleName:
        return ModuleName(*self._parts, *names)


class PyModuleType(Enum):
    NAMESPACE_PACKAGE = auto()
    REGULAR_PACKAGE = auto()
    MODULE = auto()


def _compute_py_module_type(path: Path) -> PyModuleType:
    if path.is_file():
        if path.name == "__init__.py":
            return PyModuleType.REGULAR_PACKAGE
        if path.suffix == ".py":
            return PyModuleType.MODULE
    if path.is_dir() and not (path / "__init__.py").exists():
        return PyModuleType.NAMESPACE_PACKAGE
    raise ValueError(path)


def _compute_py_module_name(package: Path, path: Path, py_module_type: PyModuleType) -> ModuleName:
    module_name = ModuleName(package.name)
    ref_path = path.parent if py_module_type is PyModuleType.REGULAR_PACKAGE else path
    if not ref_path.is_relative_to(package) or (rel_path := ref_path.relative_to(package)) == Path(
        "."
    ):
        return module_name
    return module_name.joinname(*rel_path.with_suffix("").parts)


class PyModule:
    def __init__(self, package: Path, path: Path) -> None:
        self.package: Final[Path] = package
        self.path: Final[Path] = path
        self.type: Final[PyModuleType] = _compute_py_module_type(path)
        self.name: Final[ModuleName] = _compute_py_module_name(package, path, self.type)

    def __str__(self) -> str:
        match self.type:
            case PyModuleType.NAMESPACE_PACKAGE:
                return f"{self.name}/"
            case PyModuleType.REGULAR_PACKAGE:
                return f"{self.name}.__init__"
            case PyModuleType.MODULE:
                return f"{self.name}"

    def __hash__(self) -> int:
        return hash(self.path)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PyModule):
            return NotImplemented
        return self.name == other.name

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, PyModule):
            return NotImplemented
        return self.name < other.name

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, PyModule):
            return NotImplemented
        return self.name > other.name

    def __le__(self, other: object) -> bool:
        return self < other or self == other

    def __ge__(self, other: object) -> bool:
        return self > other or self == other



#!/usr/bin/env python3

import ast
import sys
from collections.abc import Iterator, Mapping, Sequence
from dataclasses import dataclass
from pathlib import Path

from .log import logger
from .modules import ModuleName, PyModule, PyModuleType

STDLIB_OR_BUILTIN = sys.stdlib_module_names.union(sys.builtin_module_names)
ImportSTMT = ast.Import | ast.ImportFrom


@dataclass(frozen=True)
class _AbsImportStmt:
    names: tuple[str, ...]


@dataclass(frozen=True)
class _AbsImportFromStmt:
    module: str
    names: tuple[str, ...]

    def __post_init__(self) -> None:
        assert self.module


@dataclass(frozen=True)
class _RelImportFromStmt:
    level: int
    module: str
    names: tuple[str, ...]

    def __post_init__(self) -> None:
        assert self.level >= 1


def _is_import_stmt(import_stmt: str, raw_node: str) -> bool:
    # Either the args are the same or 'from a import (' vs 'from a import b'
    import_stmt_parts = [i.strip() for i in import_stmt.split()]
    raw_node_parts = [n.strip() for n in raw_node.split()]
    return import_stmt_parts == raw_node_parts or import_stmt_parts[:-1] == raw_node_parts[:-1]


def _has_ignore_comment(comments: Sequence[str]) -> bool:
    return "py-import-cycles: ignore" in [c.strip() for c in comments]


def _ignore(code_lines: Sequence[str], raw_node: str, lineno: int) -> bool:
    # === 1 ===
    # import foo.bar # linter-1: ... # py-import-cycles: ignore # linter-2: ...
    # === 2 ===
    # from foo (  # linter-1: ... # py-import-cycles: ignore # linter-2: ...
    #     bar,
    #     baz,
    #     ...
    # )
    try:
        import_stmt, *comments = code_lines[lineno - 1].split("#")
    except IndexError:
        import_stmt = ""
        comments = []

    if _is_import_stmt(import_stmt, raw_node) and _has_ignore_comment(comments):
        logger.debug("(same line) Found ignore statment for %r", raw_node)
        return True

    # === 3 ===
    # # linter-1: ... # py-import-cycles: ignore # linter-2: ...
    # import foo.bar
    try:
        comment_line = code_lines[lineno - 2]
        import_stmt = code_lines[lineno - 1]
    except IndexError:
        comment_line = ""
        import_stmt = ""

    if (
        _is_import_stmt(import_stmt, raw_node)
        and comment_line.strip().startswith("#")
        and _has_ignore_comment(comment_line.split("#")[1:])
    ):
        logger.debug("(above line) Found ignore statment for %r", raw_node)
        return True

    return False


class NodeVisitorImports(ast.NodeVisitor):
    def __init__(self, code: str) -> None:
        self._code_lines = code.split("\n")
        self._abs_import_stmts: list[_AbsImportStmt] = []
        self._abs_import_from_stmts: list[_AbsImportFromStmt] = []
        self._rel_import_from_stmts: list[_RelImportFromStmt] = []

    @property
    def abs_import_stmts(self) -> Sequence[_AbsImportStmt]:
        return self._abs_import_stmts

    @property
    def abs_import_from_stmts(self) -> Sequence[_AbsImportFromStmt]:
        return self._abs_import_from_stmts

    @property
    def rel_import_from_stmts(self) -> Sequence[_RelImportFromStmt]:
        return self._rel_import_from_stmts

    def visit_Import(self, node: ast.Import) -> None:
        if _ignore(self._code_lines, ast.unparse(node), node.lineno):
            return
        self._abs_import_stmts.append(
            _AbsImportStmt(
                tuple(a.name for a in node.names),
            )
        )

    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        if _ignore(self._code_lines, ast.unparse(node), node.lineno):
            return
        if node.level >= 1:
            self._rel_import_from_stmts.append(
                _RelImportFromStmt(
                    node.level,
                    node.module or "",
                    tuple(a.name for a in node.names),
                )
            )
        else:
            self._abs_import_from_stmts.append(
                _AbsImportFromStmt(
                    node.module or "",
                    tuple(a.name for a in node.names),
                )
            )


def _compute_py_module_from_module_name(
    py_modules_by_name: Mapping[ModuleName, PyModule], module_name: ModuleName
) -> Iterator[PyModule]:
    if not module_name.parts or module_name.parts[0] in STDLIB_OR_BUILTIN:
        return

    if module_name.parts[-1] == "*":
        # Note:
        # from a.b import *
        # - If a.b is a pkg everything from a.b.__init__.py is loaded
        # - If a.b is a mod everything from a.b.py is loaded
        # -> Getting the "right" filepath is already handled below
        module_name = module_name.parent

    if import_py_module := py_modules_by_name.get(module_name):
        yield import_py_module


def _compute_py_module_from_abs_import_stmt(
    py_modules_by_name: Mapping[ModuleName, PyModule], abs_import_stmt: _AbsImportStmt
) -> Iterator[PyModule]:
    # https://docs.python.org/3/reference/simple_stmts.html#import
    # import foo                 # foo imported and bound locally
    # import foo.bar.baz         # foo, foo.bar, and foo.bar.baz imported, foo bound locally
    # import foo.bar.baz as fbb  # foo, foo.bar, and foo.bar.baz imported, foo.bar.baz bound as fbb
    for name in abs_import_stmt.names:
        yield from _compute_py_module_from_module_name(py_modules_by_name, ModuleName(name))


def _compute_py_module_from_abs_import_from_stmt(
    py_modules_by_name: Mapping[ModuleName, PyModule],
    abs_import_from_stmt: _AbsImportFromStmt,
) -> Iterator[PyModule]:
    # https://docs.python.org/3/reference/simple_stmts.html#import
    # from foo.bar import baz    # foo, foo.bar, and foo.bar.baz imported, foo.bar.baz bound as baz
    # from foo import attr       # foo imported and foo.attr bound as attr
    anchor = ModuleName(abs_import_from_stmt.module)
    import_py_modules = [
        import_py_module
        for name in abs_import_from_stmt.names
        for import_py_module in _compute_py_module_from_module_name(
            py_modules_by_name, anchor.joinname(name)
        )
    ]
    yield from import_py_modules

    if len(import_py_modules) != len(abs_import_from_stmt.names):
        # from a.b import c, d
        # Check if c and d are modules or objects. If c or d is not a module then we have to
        # collect 'a.b'
        yield from _compute_py_module_from_module_name(py_modules_by_name, anchor)


def _compute_ref_path_from_rel_import_from_stmt(
    base_py_module: PyModule, rel_import_from_stmt: _RelImportFromStmt
) -> Path:
    # Note: PyModuleType.NAMESPACE_PACKAGE are already excluded when calling 'visit_py_module'
    if base_py_module.type is PyModuleType.MODULE:
        return base_py_module.path.parents[rel_import_from_stmt.level - 1]
    # PyModuleType.REGULAR_PACKAGE
    if rel_import_from_stmt.level == 1:
        return base_py_module.path.parent
    return base_py_module.path.parents[rel_import_from_stmt.level - 2]


def _compute_py_module_from_rel_import_from_stmt(base_py_module: PyModule, path: Path) -> PyModule:
    if (module_file_path := path.with_suffix(".py")).exists():
        return PyModule(package=base_py_module.package, path=module_file_path)
    if (init_file_path := path / "__init__.py").exists():
        return PyModule(package=base_py_module.package, path=init_file_path)
    if path.is_dir():
        return PyModule(package=base_py_module.package, path=path)
    raise ValueError(path)


def _compute_py_modules_from_rel_import_from_stmt(
    base_py_module: PyModule, rel_import_from_stmt: _RelImportFromStmt
) -> Iterator[PyModule]:
    ref_path = _compute_ref_path_from_rel_import_from_stmt(base_py_module, rel_import_from_stmt)

    if rel_import_from_stmt.module:
        ref_path = ref_path.joinpath(*ModuleName(rel_import_from_stmt.module).parts)
        try:
            yield _compute_py_module_from_rel_import_from_stmt(base_py_module, ref_path)
        except ValueError:
            logger.debug("Cannot make py module from %s", ref_path)

    for name in rel_import_from_stmt.names:
        try:
            yield _compute_py_module_from_rel_import_from_stmt(
                base_py_module, ref_path.joinpath(name)
            )
        except ValueError:
            logger.debug("Cannot make py module from %s", ref_path)


def _is_valid(base_py_module: PyModule, import_py_module: PyModule) -> bool:
    if base_py_module == import_py_module:
        return False
    if (
        base_py_module.type is PyModuleType.MODULE
        and import_py_module.type is PyModuleType.REGULAR_PACKAGE
        and base_py_module.path.parent == import_py_module.path.parent
    ):
        # Do not take 'foo.bar.baz imports foo.bar.__init__' into account
        return False
    return True


def visit_py_module(
    py_modules_by_name: Mapping[ModuleName, PyModule], base_py_module: PyModule
) -> Iterator[PyModule]:
    if base_py_module.type is PyModuleType.NAMESPACE_PACKAGE:
        return

    try:
        with open(base_py_module.path, encoding="utf-8") as f:
            code = f.read()
    except UnicodeDecodeError as e:
        logger.debug("Cannot read python file %s: %s", base_py_module.path, e)
        return

    try:
        tree = ast.parse(code)
    except SyntaxError as e:
        logger.debug("Cannot visit python file %s: %s", base_py_module.path, e)
        return

    visitor = NodeVisitorImports(code)
    visitor.visit(tree)

    for abs_import_stmt in visitor.abs_import_stmts:
        for import_py_module in _compute_py_module_from_abs_import_stmt(
            py_modules_by_name, abs_import_stmt
        ):
            if _is_valid(base_py_module, import_py_module):
                yield import_py_module

    for abs_import_from_stmt in visitor.abs_import_from_stmts:
        for import_py_module in _compute_py_module_from_abs_import_from_stmt(
            py_modules_by_name, abs_import_from_stmt
        ):
            if _is_valid(base_py_module, import_py_module):
                yield import_py_module

    for rel_import_from_stmt in visitor.rel_import_from_stmts:
        for import_py_module in _compute_py_modules_from_rel_import_from_stmt(
            base_py_module, rel_import_from_stmt
        ):
            if _is_valid(base_py_module, import_py_module):
                yield import_py_module



#!/usr/bin/env python3

import os
from collections.abc import Mapping, Sequence
from typing import Any

from .modules import PyModule


def make_sarif(
    sorted_cycles: Sequence[tuple[PyModule, ...]],
    version: str,
) -> Mapping[str, Any]:
    results = []
    for cycle in sorted_cycles:
        cycle_str = " -> ".join(str(m) for m in cycle) + " -> " + str(cycle[0])

        locations = [
            {
                "physicalLocation": {
                    "artifactLocation": {"uri": os.path.relpath(cycle[0].path)},
                    "region": {"startLine": 1, "startColumn": 1},
                },
            },
        ]

        related_locations = [
            {
                "id": idx,
                "physicalLocation": {
                    "artifactLocation": {"uri": os.path.relpath(m.path)},
                    "region": {"startLine": 1, "startColumn": 1},
                },
            }
            for idx, m in enumerate(cycle[1:], start=1)
        ]

        result: dict = {
            "ruleId": "import-cycle",
            "level": "warning",
            "message": {"text": f"Import cycle: {cycle_str}"},
            "locations": locations,
        }
        if related_locations:
            result["relatedLocations"] = related_locations

        results.append(result)

    return {
        "$schema": (
            # pylint: disable=line-too-long
            "https://raw.githubusercontent.com/oasis-tcs/sarif-spec/main/sarif-2.1/schema/sarif-schema-2.1.0.json"
        ),
        "version": "2.1.0",
        "runs": [
            {
                "tool": {
                    "driver": {
                        "name": "py-import-cycles",
                        "version": version,
                        "rules": [
                            {
                                "id": "import-cycle",
                                "shortDescription": {
                                    "text": "Import cycle detected",
                                },
                            },
                        ],
                    },
                },
                "results": results,
            },
        ],
    }



#!/usr/bin/env python3

import argparse
import json
import logging
import sys
import time
from collections import Counter
from collections.abc import Iterator, Mapping, Sequence
from pathlib import Path
from typing import Literal

from . import __version__
from .cycles import detect_cycles
from .files import get_outputs_file_paths, parse_files, scan_packages
from .graphs import make_graph
from .log import logger, setup_logging
from .modules import PyModule
from .sarif import make_sarif
from .visitors import visit_py_module


def _parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="""description:
  Detect import cycles in Python projects.

  py-import-cycles is conceived for having an indication whether Python packages may have
  structural weak points and does not take any Python module finder or loader mechanisms into
  account.

  Single import statements can be ignored via '# py-import-cycles: ignore':

    import a  # py-import-cycles: ignore

    # py-import-cycles: ignore
    import b
""",
        formatter_class=argparse.RawTextHelpFormatter,
        allow_abbrev=False,
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=__version__,
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="show additional information for debug purposes",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="show cycles if some are found",
    )
    parser.add_argument(
        "--outputs-folder",
        help="path to outputs folder. If not set $HOME/.local/py-import-cycles/outputs/ is used",
    )
    parser.add_argument(
        "--outputs-filename",
        help="outputs filename. If not set the current timestamp is used",
    )
    parser.add_argument(
        "--packages",
        nargs="+",
        help="collect Python files from top-level packages",
        default=[],
    )
    parser.add_argument(
        "--files",
        nargs="+",
        help="""Directly pass files for cycle analysis.
Note that a given file must match one of the following patterns:
  - /path/to/repo::pkg/module.py (absolute)
  - relpath/to/repo::pkg/module.py (relative)
In the second case the relpath will be extended with the current working directory.""",
        default=[],
    )
    parser.add_argument(
        "--strategy",
        choices=["dfs", "tarjan", "johnson"],
        default="dfs",
        help="path-based strong component algorithm",
    )
    parser.add_argument(
        "--threshold",
        type=int,
        default=0,
        help="Tolerate a certain number of cycles, ie. an upper threshold.",
    )
    parser.add_argument(
        "--graph",
        action="store_true",
        help="create graphical representation",
    )
    parser.add_argument(
        "--stats",
        action="store_true",
        help="show some statistics",
    )
    parser.add_argument(
        "--sarif",
        help="write SARIF output to the given file, or '-' for stdout",
    )

    return parser.parse_args()


def main() -> int:
    now = time.time()
    stats: Counter[Literal["num_of_modules", "num_of_modules_with_imports", "num_of_imports"]] = (
        Counter()
    )

    args = _parse_arguments()

    packages = [pp for p in args.packages if (pp := Path(p)).is_dir()]

    outputs_filepaths = get_outputs_file_paths(
        Path(args.outputs_folder) if args.outputs_folder else None,
        args.outputs_filename or "",
    )

    setup_logging(outputs_filepaths.log, args.debug)

    logger.info("Scan packages")
    py_modules = list(scan_packages(packages))
    if args.files:
        logger.info("Parse files")
        py_modules.extend(parse_files(args.files))

    if args.stats:
        stats["num_of_modules"] = len(py_modules)

    logger.info("Visit and compute imports of py modules")
    py_modules_by_name = {p.name: p for p in py_modules}

    imports_by_py_module = {
        py_module: imports
        for py_module in py_modules
        if (
            imports := sorted(
                frozenset(visit_py_module(py_modules_by_name, py_module)),
                key=lambda m: tuple(m.name.parts),
                reverse=True,
            )
        )
    }

    if args.stats:
        stats["num_of_imports"] = sum(len(ims) for ims in imports_by_py_module.values())
        stats["num_of_modules_with_imports"] = len(imports_by_py_module)

    if _debug():
        logger.debug(
            "Imports of py modules:\n%s",
            "\n".join(_make_readable_imports_by_py_module(imports_by_py_module)),
        )

    logger.info("Detect import cycles with strategy %s", args.strategy)
    unsorted_cycles = set(detect_cycles(args.strategy, imports_by_py_module))

    logger.info("Sort import cycles")
    sorted_cycles = sorted(unsorted_cycles, key=lambda t: (len(t), t[0].name))

    sys.stderr.write(f"Found {len(sorted_cycles)} import cycles\n")
    _log_or_show_cycles(args.verbose, sorted_cycles)

    if args.sarif:
        _write_sarif(args.sarif, sorted_cycles)

    if args.graph:
        logger.info("Close cycle and make graph")
        make_graph(
            outputs_filepaths.graph,
            args.strategy,
            [((cycle[-1],) + cycle) for cycle in sorted_cycles],
        )

    if args.stats:
        _show_stats(time.time() - now, stats)

    return len(unsorted_cycles) > args.threshold


#   .--helper--------------------------------------------------------------.
#   |                    _          _                                      |
#   |                   | |__   ___| |_ __   ___ _ __                      |
#   |                   | '_ \ / _ \ | '_ \ / _ \ '__|                     |
#   |                   | | | |  __/ | |_) |  __/ |                        |
#   |                   |_| |_|\___|_| .__/ \___|_|                        |
#   |                                |_|                                   |
#   '----------------------------------------------------------------------'


def _make_readable_imports_by_py_module(
    imports_by_py_module: Mapping[PyModule, Sequence[PyModule]],
) -> Iterator[str]:
    for ibm, ms in imports_by_py_module.items():
        if ms:
            yield f"  {str(ibm)} imports:"
            yield from (f"    {str(m)}" for m in ms)


def _make_readable_cycles(
    sorted_cycles: Sequence[tuple[PyModule, ...]],
) -> Iterator[str]:
    for nr, ic in enumerate(sorted_cycles, start=1):
        if ic:
            yield f"  Cycle {nr}:"
            yield f"    {str(ic[0])}"
            yield from (f"    > {str(i)} " for i in ic[1:])


def _log_or_show_cycles(
    verbose: bool,
    sorted_cycles: Sequence[tuple[PyModule, ...]],
) -> None:
    if verbose:
        for line in _make_readable_cycles(sorted_cycles):
            sys.stderr.write(f"{line}\n")

    if _debug():
        logger.debug(
            "Import cycles:\n%s",
            "\n".join(_make_readable_cycles(sorted_cycles)),
        )


def _show_stats(
    duration: float,
    stats: Counter[Literal["num_of_modules", "num_of_modules_with_imports", "num_of_imports"]],
) -> None:
    logger.info("Duration: %ss", f"{duration:.2f}")
    logger.info("Number of found modules: %s", stats["num_of_modules"])
    logger.info("Number of imports: %s", stats["num_of_imports"])
    logger.info("Number of modules with imports: %s", stats["num_of_modules_with_imports"])


def _write_sarif(
    path: str,
    sorted_cycles: Sequence[tuple[PyModule, ...]],
) -> None:
    sarif = json.dumps(make_sarif(sorted_cycles, __version__), indent=2) + "\n"
    if path == "-":
        sys.stdout.write(sarif)
    else:
        Path(path).write_text(sarif, encoding="utf-8")


def _debug() -> bool:
    return logger.level == logging.DEBUG
